import { navigator, window, document } from 'global';
import React from 'react';
import ReactDOM from 'react-dom';
import deprecate from 'util-deprecate';
import AnsiToHtml from 'ansi-to-html';

import addons from '@storybook/addons';
import createChannel from '@storybook/channel-postmessage';
import { ClientApi, StoryStore, ConfigApi } from '@storybook/client-api';
import { toId, storyNameFromExport, isExportStory } from '@storybook/csf';
import { logger } from '@storybook/client-logger';
import Events from '@storybook/core-events';

import { initializePath, setPath } from './url';
import { NoDocs } from './NoDocs';

const ansiConverter = new AnsiToHtml({
  escapeXML: true,
});

const classes = {
  MAIN: 'sb-show-main',
  NOPREVIEW: 'sb-show-nopreview',
  ERROR: 'sb-show-errordisplay',
};

function showMain() {
  document.body.classList.remove(classes.NOPREVIEW);
  document.body.classList.remove(classes.ERROR);

  document.body.classList.add(classes.MAIN);
}

function showNopreview() {
  document.body.classList.remove(classes.MAIN);
  document.body.classList.remove(classes.ERROR);

  document.body.classList.add(classes.NOPREVIEW);
}

function showErrorDisplay({ message = '', stack = '' }) {
  document.getElementById('error-message').innerHTML = ansiConverter.toHtml(message);
  document.getElementById('error-stack').innerHTML = ansiConverter.toHtml(stack);

  document.body.classList.remove(classes.MAIN);
  document.body.classList.remove(classes.NOPREVIEW);

  document.body.classList.add(classes.ERROR);
}

// showError is used by the various app layers to inform the user they have done something
// wrong -- for instance returned the wrong thing from a story
function showError({ title, description }) {
  addons.getChannel().emit(Events.STORY_ERRORED, { title, description });
  showErrorDisplay({
    message: title,
    stack: description,
  });
}

// showException is used if we fail to render the story and it is uncaught by the app layer
function showException(exception) {
  addons.getChannel().emit(Events.STORY_THREW_EXCEPTION, exception);
  showErrorDisplay(exception);

  // Log the stack to the console. So, user could check the source code.
  logger.error(exception);
}

const isBrowser =
  navigator &&
  navigator.userAgent &&
  navigator.userAgent !== 'storyshots' &&
  !(navigator.userAgent.indexOf('Node.js') > -1) &&
  !(navigator.userAgent.indexOf('jsdom') > -1);

export const getContext = (() => {
  let cache;
  return decorateStory => {
    if (cache) {
      return cache;
    }
    let channel = null;
    if (isBrowser) {
      try {
        channel = addons.getChannel();
      } catch (e) {
        channel = createChannel({ page: 'preview' });
        addons.setChannel(channel);
      }
    }
    let storyStore;
    let clientApi;
    if (typeof window !== 'undefined' && window.__STORYBOOK_CLIENT_API__) {
      clientApi = window.__STORYBOOK_CLIENT_API__;
      // eslint-disable-next-line no-underscore-dangle
      storyStore = clientApi._storyStore;
    } else {
      storyStore = new StoryStore({ channel });
      clientApi = new ClientApi({ storyStore, decorateStory });
    }
    const { clearDecorators } = clientApi;
    const configApi = new ConfigApi({ clearDecorators, storyStore, channel, clientApi });

    return {
      configApi,
      storyStore,
      channel,
      clientApi,
      showMain,
      showError,
      showException,
    };
  };
})();

function focusInInput(event) {
  return (
    /input|textarea/i.test(event.target.tagName) ||
    event.target.getAttribute('contenteditable') !== null
  );
}

const applyLayout = (() => {
  let previousStyles;

  return layout => {
    const layouts = {
      centered: `
        display: flex;
        justify-content: center;
        align-items: center;
        min-height: 100vh;
        margin: 0;
        padding: 1rem;
        box-sizing: border-box;
      `,
      fullscreen: `
        margin: 0;
      `,
      padded: `
        margin: 0;
        padding: 1rem;
      `,
    };
    const styles = layouts[layout] || layouts.padded;

    if (styles !== previousStyles) {
      document.body.style = styles;
      previousStyles = styles;
    }
  };
})();

export default function start(render, { decorateStory } = {}) {
  const context = getContext(decorateStory);

  const { clientApi, channel, configApi, storyStore } = context;
  // Provide access to external scripts if `window` is defined.
  // NOTE this is different to isBrowser, primarily for the JSDOM use case
  let previousKind = '';
  let previousStory = '';
  let previousRevision = -1;
  let previousViewMode = '';
  let previousId = null;

  const renderMain = forceRender => {
    const revision = storyStore.getRevision();
    const loadError = storyStore.getError();
    const { storyId, viewMode: urlViewMode } = storyStore.getSelection();

    const data = storyStore.fromId(storyId);

    const { kind, name, getDecorated, id, parameters = {}, error } = data || {};

    const { docsOnly, layout } = parameters;
    applyLayout(layout);

    const viewMode = docsOnly ? 'docs' : urlViewMode;

    const renderContext = {
      ...context,
      ...data,
      selectedKind: kind,
      selectedStory: name,
      parameters,
      forceRender,
    };

    if (loadError || error) {
      showErrorDisplay(loadError || error);
      return;
    }

    // Render story only if selectedKind or selectedStory have changed.
    // However, we DO want the story to re-render if the store itself has changed
    // (which happens at the moment when HMR occurs)
    if (
      !forceRender &&
      revision === previousRevision &&
      viewMode === previousViewMode &&
      kind === previousKind &&
      name === previousStory
    ) {
      addons.getChannel().emit(Events.STORY_UNCHANGED, {
        id,
        revision,
        kind,
        name,
        viewMode,
      });
      return;
    }

    if (!forceRender && previousKind && previousStory && revision === previousRevision) {
      addons.getChannel().emit(Events.STORY_CHANGED, id);
    }

    switch (previousViewMode) {
      case 'docs':
        if (previousKind != null && (kind !== previousKind || viewMode !== previousViewMode)) {
          storyStore.cleanHooksForKind(previousKind);
          ReactDOM.unmountComponentAtNode(document.getElementById('docs-root'));
        }
        break;
      case 'story':
      default:
        if (previousId != null && (id !== previousId || viewMode !== previousViewMode)) {
          storyStore.cleanHooks(previousId);
          ReactDOM.unmountComponentAtNode(document.getElementById('root'));
        }
    }

    // Docs view renders into a different root ID to avoid conflicts
    // with the user's view layer. Therefore we need to clean up whenever
    // we transition between view modes
    if (viewMode !== previousViewMode) {
      switch (viewMode) {
        case 'docs': {
          showMain();
          document.getElementById('root').setAttribute('hidden', true);
          document.getElementById('docs-root').removeAttribute('hidden');
          break;
        }
        case 'story':
        default: {
          if (previousViewMode === 'docs') {
            document.getElementById('docs-root').setAttribute('hidden', true);
            document.getElementById('root').removeAttribute('hidden');
          }
        }
      }
    }
    // Given a cleaned up state, render the appropriate view mode
    switch (viewMode) {
      case 'docs': {
        const docs = parameters.docs || {};
        const DocsContainer = docs.container || (({ children }) => <>{children}</>);
        const Page = docs.page || NoDocs;
        ReactDOM.render(
          <DocsContainer context={renderContext}>
            <Page />
          </DocsContainer>,
          document.getElementById('docs-root'),
          () => addons.getChannel().emit(Events.DOCS_RENDERED, kind)
        );
        break;
      }
      case 'story':
      default: {
        if (getDecorated) {
          (async () => {
            try {
              await render(renderContext);
              addons.getChannel().emit(Events.STORY_RENDERED, id);
            } catch (ex) {
              showException(ex);
            }
          })();
        } else {
          showNopreview();
          addons.getChannel().emit(Events.STORY_MISSING, id);
        }
        break;
      }
    }

    previousRevision = revision;
    previousKind = kind;
    previousStory = name;
    previousViewMode = viewMode;
    previousId = id;

    if (!forceRender && viewMode !== 'docs') {
      document.documentElement.scrollTop = 0;
      document.documentElement.scrollLeft = 0;
    }
  };

  // initialize the UI
  const renderUI = forceRender => {
    if (isBrowser) {
      try {
        renderMain(forceRender);
      } catch (ex) {
        showException(ex);
      }
    }
  };

  const forceReRender = () => renderUI(true);

  // channel can be null in NodeJS
  if (isBrowser) {
    const deprecatedToId = deprecate(
      toId,
      `Passing name+kind to the SET_CURRENT_STORY event is deprecated, use a storyId instead`
    );

    channel.on(Events.FORCE_RE_RENDER, forceReRender);
    channel.on(Events.SET_CURRENT_STORY, ({ storyId: inputStoryId, name, kind, viewMode }) => {
      let storyId = inputStoryId;
      // For backwards compatibility
      if (!storyId) {
        if (!name || !kind) {
          throw new Error('You should pass `storyId` into SET_CURRENT_STORY');
        }
        storyId = deprecatedToId(kind, name);
      }

      storyStore.setSelection({ storyId, viewMode });
      setPath({ storyId, viewMode });
    });

    // Handle keyboard shortcuts
    window.onkeydown = event => {
      if (!focusInInput(event)) {
        // We have to pick off the keys of the event that we need on the other side
        const { altKey, ctrlKey, metaKey, shiftKey, key, code, keyCode } = event;
        channel.emit(Events.PREVIEW_KEYDOWN, {
          event: { altKey, ctrlKey, metaKey, shiftKey, key, code, keyCode },
        });
      }
    };
  }

  storyStore.on(Events.STORY_INIT, () => {
    const { storyId, viewMode } = initializePath(storyStore);
    storyStore.setSelection({ storyId, viewMode });
  });

  storyStore.on(Events.STORY_RENDER, renderUI);

  if (typeof window !== 'undefined') {
    window.__STORYBOOK_CLIENT_API__ = clientApi;
    window.__STORYBOOK_STORY_STORE__ = storyStore;
    window.__STORYBOOK_ADDONS_CHANNEL__ = channel; // may not be defined
  }

  let previousExports = new Map();
  const loadStories = (loadable, framework) => () => {
    let reqs = null;
    if (Array.isArray(loadable)) {
      reqs = loadable;
    } else if (loadable.keys) {
      reqs = [loadable];
    }

    let currentExports = new Map();
    if (reqs) {
      reqs.forEach(req => {
        req.keys().forEach(filename => {
          const fileExports = req(filename);
          currentExports.set(
            fileExports,
            typeof req.resolve === 'function' ? req.resolve(filename) : null
          );
        });
      });
    } else {
      const exported = loadable();
      if (Array.isArray(exported) && exported.every(obj => obj.default != null)) {
        currentExports = new Map(exported.map(fileExports => [fileExports, null]));
      } else if (exported) {
        logger.warn(
          `Loader function passed to 'configure' should return void or an array of module exports that all contain a 'default' export. Received: ${JSON.stringify(
            exported
          )}`
        );
      }
    }

    const removed = [...previousExports.keys()].filter(exp => !currentExports.has(exp));
    removed.forEach(exp => {
      if (exp.default) {
        storyStore.removeStoryKind(exp.default.title);
      }
    });
    if (removed.length > 0) {
      storyStore.incrementRevision();
    }

    const added = [...currentExports.keys()].filter(exp => !previousExports.has(exp));
    added.forEach(fileExports => {
      // An old-style story file
      if (!fileExports.default) {
        return;
      }

      if (!fileExports.default.title) {
        throw new Error(
          `Unexpected default export without title: ${JSON.stringify(fileExports.default)}`
        );
      }

      const { default: meta, __namedExportsOrder, ...namedExports } = fileExports;
      let exports = namedExports;

      // prefer a user/loader provided `__namedExportsOrder` array if supplied
      // we do this as es module exports are always ordered alphabetically
      // see https://github.com/storybookjs/storybook/issues/9136
      if (Array.isArray(__namedExportsOrder)) {
        exports = {};
        __namedExportsOrder.forEach(name => {
          if (namedExports[name]) {
            exports[name] = namedExports[name];
          }
        });
      }

      const {
        title: kindName,
        id: componentId,
        parameters: params,
        decorators: decos,
        component,
        subcomponents,
      } = meta;
      // We pass true here to avoid the warning about HMR. It's cool clientApi, we got this
      const kind = clientApi.storiesOf(kindName, true);

      // we should always have a framework, rest optional
      kind.addParameters({
        framework,
        component,
        subcomponents,
        fileName: currentExports.get(fileExports),
        ...params,
      });

      (decos || []).forEach(decorator => {
        kind.addDecorator(decorator);
      });

      Object.keys(exports).forEach(key => {
        if (isExportStory(key, meta)) {
          const storyFn = exports[key];
          const { name, parameters, decorators } = storyFn.story || {};
          if (parameters && parameters.decorators) {
            deprecate(() => {},
            `${kindName} => ${name || key}: story.parameters.decorators is deprecated; use story.decorators instead.`)();
          }
          const decoratorParams = decorators ? { decorators } : null;
          const exportName = storyNameFromExport(key);
          const idParams = { __id: toId(componentId || kindName, exportName) };
          kind.add(name || exportName, storyFn, {
            ...parameters,
            ...decoratorParams,
            ...idParams,
          });
        }
      });
    });
    previousExports = currentExports;
  };

  let loaded = false;
  /**
   * Load a collection of stories. If it has a default export, assume that it is a module-style
   * file and process its named exports as stories. If not, assume it's an old-style
   * storiesof file and require it.
   *
   * @param {*} loadable a require.context `req`, an array of `req`s, or a loader function that returns void or an array of exports
   * @param {*} m - ES module object for hot-module-reloading (HMR)
   * @param {*} framework - name of framework in use, e.g. "react"
   */
  const configure = (loadable, m, framework) => {
    if (typeof m === 'string') {
      throw new Error(
        `Invalid module '${m}'. Did you forget to pass \`module\` as the second argument to \`configure\`"?`
      );
    }
    if (m && m.hot && m.hot.dispose) {
      ({ previousExports = new Map() } = m.hot.data || {});

      m.hot.dispose(data => {
        loaded = false;
        // eslint-disable-next-line no-param-reassign
        data.previousExports = previousExports;
      });
    }
    if (loaded) {
      logger.warn('Unexpected loaded state. Did you call `load` twice?');
    }
    loaded = true;
    configApi.configure(loadStories(loadable, framework), m);
  };

  return { configure, context, clientApi, configApi, forceReRender };
}
